# Generated by Django 5.0.10 on 2025-03-11 16:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0085_alter_agent_output_modes"),
    ]

    operations = [
        migrations.AlterField(
            model_name="texttoimagemodelconfig",
            name="model_type",
            field=models.CharField(
                choices=[
                    ("openai", "Openai"),
                    ("stability-ai", "Stabilityai"),
                    ("replicate", "Replicate"),
                    ("google", "Google"),
                ],
                default="openai",
                max_length=200,
            ),
        ),
    ]
