# Generated by Django 5.0.9 on 2024-12-14 18:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0077_chatmodel_alter_agent_chat_model_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="khojuser",
            name="email_verification_code_expiry",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]
