# Generated by Django 5.1.14 on 2025-11-27 01:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0096_mcpserver"),
    ]

    operations = [
        migrations.AddField(
            model_name="serverchatsettings",
            name="priority",
            field=models.IntegerField(
                blank=True,
                default=None,
                help_text="Priority of the server chat settings. Lower numbers run first.",
                null=True,
                unique=True,
            ),
        ),
    ]
