# Generated by Django 5.1.10 on 2025-07-19 21:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0091_chatmodel_friendly_name_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="chatmodel",
            name="model_type",
            field=models.CharField(
                choices=[("openai", "Openai"), ("anthropic", "Anthropic"), ("google", "Google")],
                default="google",
                max_length=200,
            ),
        ),
        migrations.AlterField(
            model_name="chatmodel",
            name="name",
            field=models.CharField(default="gemini-2.5-flash", max_length=200),
        ),
        migrations.AlterField(
            model_name="speechtotextmodeloptions",
            name="model_name",
            field=models.CharField(default="whisper-1", max_length=200),
        ),
        migrations.AlterField(
            model_name="speechtotextmodeloptions",
            name="model_type",
            field=models.CharField(choices=[("openai", "Openai")], default="openai", max_length=200),
        ),
    ]
