"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5142],{75142:(e,a,r)=>{r.d(a,{diagram:()=>o});var t=r(68042),n=r(80962),s=r(20316),d=r(34290),i={parse:(0,s.K2)(async e=>{let a=await (0,d.qg)("info",e);s.Rm.debug(a)},"parse")},p={version:"11.13.0"},g=(0,s.K2)(()=>p.version,"getVersion"),o={parser:i,db:{getVersion:g},renderer:{draw:(0,s.K2)((e,a,r)=>{s.Rm.debug("rendering info diagram\n"+e);let d=(0,t.D)(a);(0,n.a$)(d,100,400,!0),d.append("g").append("text").attr("x",100).attr("y",40).attr("class","version").attr("font-size",32).style("text-anchor","middle").text(`v${r}`)},"draw")}}}}]);