# Generated by Django 5.1.10 on 2025-07-25 23:30

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0092_alter_chatmodel_model_type_alter_chatmodel_name_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="localorgconfig",
            name="user",
        ),
        migrations.RemoveField(
            model_name="localpdfconfig",
            name="user",
        ),
        migrations.RemoveField(
            model_name="localplaintextconfig",
            name="user",
        ),
        migrations.DeleteModel(
            name="LocalMarkdownConfig",
        ),
        migrations.DeleteModel(
            name="LocalOrgConfig",
        ),
        migrations.DeleteModel(
            name="LocalPdfConfig",
        ),
        migrations.DeleteModel(
            name="LocalPlaintextConfig",
        ),
    ]
