# Generated by Django 5.1.14 on 2025-11-13 20:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0095_alter_webscraper_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="McpServer",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=50, unique=True)),
                ("path", models.CharField(max_length=200, unique=True)),
                ("api_key", models.CharField(blank=True, max_length=4000, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
