# Generated by Django 5.1.10 on 2025-08-26 20:47

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0093_remove_localorgconfig_user_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_free_deep",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_free_deep",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_free_fast",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_free_fast",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_paid_deep",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_paid_deep",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_paid_fast",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_paid_fast",
                to="database.chatmodel",
            ),
        ),
    ]
