# Generated by Django 5.1.14 on 2025-11-12 19:25

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0094_serverchatsettings_think_free_deep_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webscraper",
            name="type",
            field=models.CharField(
                choices=[
                    ("Firecrawl", "Firecrawl"),
                    ("Olostep", "Olostep"),
                    ("Exa", "Exa"),
                    ("Direct", "Direct"),
                ],
                default="Direct",
                max_length=20,
            ),
        ),
    ]
