# Generated by Django 5.0.10 on 2025-01-21 18:42
from typing import List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0081_agent_is_hidden_chatmodel_description_and_more"),
        ("database", "0081_merge_20250120_1633"),
    ]

    operations: List[str] = []
