# Generated by Django 5.1.14 on 2025-11-30 00:47

from django.db import migrations, models


def delete_stability_configs(apps, schema_editor):
    TextToImageModelConfig = apps.get_model("database", "TextToImageModelConfig")
    TextToImageModelConfig.objects.filter(model_type="stability-ai").delete()


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0097_serverchatsettings_priority"),
    ]

    operations = [
        migrations.RunPython(delete_stability_configs, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="texttoimagemodelconfig",
            name="model_type",
            field=models.CharField(
                choices=[
                    ("openai", "Openai"),
                    ("replicate", "Replicate"),
                    ("google", "Google"),
                ],
                default="openai",
                max_length=200,
            ),
        ),
    ]
