# Generated by Django 5.1.8 on 2025-04-18 15:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0088_ratelimitrecord"),
    ]

    operations = [
        migrations.AddField(
            model_name="chatmodel",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="speechtotextmodeloptions",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="texttoimagemodelconfig",
            name="price_tier",
            field=models.CharField(choices=[("free", "Free"), ("standard", "Standard")], default="free", max_length=20),
        ),
        migrations.AddField(
            model_name="voicemodeloption",
            name="price_tier",
            field=models.CharField(
                choices=[("free", "Free"), ("standard", "Standard")], default="standard", max_length=20
            ),
        ),
    ]
